@@ -19,7 +19,9 @@ use Cake\Collection\Collection;
 use Cake\Database\Expression\FieldInterface;
 use Cake\Database\Expression\QueryExpression;
 use Cake\Datasource\EntityInterface;
+use Cake\Datasource\QueryInterface;
 use Cake\ORM\Association;
+use Cake\ORM\Association\DependentDeleteHelper;
 use Cake\ORM\Association\Loader\SelectLoader;
 use Cake\ORM\Table;
 use InvalidArgumentException;
@@ -34,8 +36,6 @@ use Traversable;
 class HasMany extends Association
 {
 
-    use DependentDeleteTrait;
-
     /**
      * Order in which target records should be returned
      *
@@ -48,7 +48,7 @@ class HasMany extends Association
      *
      * @var string
      */
-    protected $_joinType = 'INNER';
+    protected $_joinType = QueryInterface::JOIN_TYPE_INNER;
 
     /**
      * The strategy name to be used to fetch associated records.
@@ -110,7 +110,7 @@ class HasMany extends Association
      */
     public function setSaveStrategy($strategy)
     {
-        if (!in_array($strategy, [self::SAVE_APPEND, self::SAVE_REPLACE])) {
+        if (!in_array($strategy, [self::SAVE_APPEND, self::SAVE_REPLACE], true)) {
             $msg = sprintf('Invalid save strategy "%s"', $strategy);
             throw new InvalidArgumentException($msg);
         }
@@ -141,6 +141,10 @@ class HasMany extends Association
      */
     public function saveStrategy($strategy = null)
     {
+        deprecationWarning(
+            'HasMany::saveStrategy() is deprecated. ' .
+            'Use setSaveStrategy()/getSaveStrategy() instead.'
+        );
         if ($strategy !== null) {
             $this->setSaveStrategy($strategy);
         }
@@ -155,8 +159,7 @@ class HasMany extends Association
      * `$options`
      *
      * @param \Cake\Datasource\EntityInterface $entity an entity from the source table
-     * @param array|\ArrayObject $options options to be passed to the save method in
-     * the target table
+     * @param array $options options to be passed to the save method in the target table
      * @return bool|\Cake\Datasource\EntityInterface false if $entity could not be saved, otherwise it returns
      * the saved entity
      * @see \Cake\ORM\Table::save()
@@ -165,63 +168,94 @@ class HasMany extends Association
     public function saveAssociated(EntityInterface $entity, array $options = [])
     {
         $targetEntities = $entity->get($this->getProperty());
-        if (empty($targetEntities) && $this->_saveStrategy !== self::SAVE_REPLACE) {
-            return $entity;
+
+        $isEmpty = in_array($targetEntities, [null, [], '', false], true);
+        if ($isEmpty) {
+            if ($entity->isNew() ||
+                $this->getSaveStrategy() !== self::SAVE_REPLACE
+            ) {
+                return $entity;
+            }
+
+            $targetEntities = [];
         }
 
-        if (!is_array($targetEntities) && !($targetEntities instanceof Traversable)) {
+        if (!is_array($targetEntities) &&
+            !($targetEntities instanceof Traversable)
+        ) {
             $name = $this->getProperty();
             $message = sprintf('Could not save %s, it cannot be traversed', $name);
             throw new InvalidArgumentException($message);
         }
 
-        $foreignKey = (array)$this->getForeignKey();
-        $properties = array_combine(
-            $foreignKey,
+        $foreignKeyReference = array_combine(
+            (array)$this->getForeignKey(),
             $entity->extract((array)$this->getBindingKey())
         );
-        $target = $this->getTarget();
-        $original = $targetEntities;
+
         $options['_sourceTable'] = $this->getSource();
 
-        $unlinkSuccessful = null;
-        if ($this->_saveStrategy === self::SAVE_REPLACE) {
-            $unlinkSuccessful = $this->_unlinkAssociated($properties, $entity, $target, $targetEntities, $options);
-        }
-
-        if ($unlinkSuccessful === false) {
+        if ($this->_saveStrategy === self::SAVE_REPLACE &&
+            !$this->_unlinkAssociated($foreignKeyReference, $entity, $this->getTarget(), $targetEntities, $options)
+        ) {
             return false;
         }
 
-        foreach ($targetEntities as $k => $targetEntity) {
-            if (!($targetEntity instanceof EntityInterface)) {
+        if (!$this->_saveTarget($foreignKeyReference, $entity, $targetEntities, $options)) {
+            return false;
+        }
+
+        return $entity;
+    }
+
+    /**
+     * Persists each of the entities into the target table and creates links between
+     * the parent entity and each one of the saved target entities.
+     *
+     * @param array $foreignKeyReference The foreign key reference defining the link between the
+     * target entity, and the parent entity.
+     * @param \Cake\Datasource\EntityInterface $parentEntity The source entity containing the target
+     * entities to be saved.
+     * @param array|\Traversable $entities list of entities to persist in target table and to
+     * link to the parent entity
+     * @param array $options list of options accepted by `Table::save()`.
+     * @return bool `true` on success, `false` otherwise.
+     */
+    protected function _saveTarget(array $foreignKeyReference, EntityInterface $parentEntity, $entities, array $options)
+    {
+        $foreignKey = array_keys($foreignKeyReference);
+        $table = $this->getTarget();
+        $original = $entities;
+
+        foreach ($entities as $k => $entity) {
+            if (!($entity instanceof EntityInterface)) {
                 break;
             }
 
             if (!empty($options['atomic'])) {
-                $targetEntity = clone $targetEntity;
+                $entity = clone $entity;
             }
 
-            if ($properties !== $targetEntity->extract($foreignKey)) {
-                $targetEntity->set($properties, ['guard' => false]);
+            if ($foreignKeyReference !== $entity->extract($foreignKey)) {
+                $entity->set($foreignKeyReference, ['guard' => false]);
             }
 
-            if ($target->save($targetEntity, $options)) {
-                $targetEntities[$k] = $targetEntity;
+            if ($table->save($entity, $options)) {
+                $entities[$k] = $entity;
                 continue;
             }
 
             if (!empty($options['atomic'])) {
-                $original[$k]->errors($targetEntity->errors());
+                $original[$k]->setErrors($entity->getErrors());
                 $entity->set($this->getProperty(), $original);
 
                 return false;
             }
         }
 
-        $entity->set($this->getProperty(), $targetEntities);
+        $parentEntity->set($this->getProperty(), $entities);
 
-        return $entity;
+        return true;
     }
 
     /**
@@ -274,7 +308,7 @@ class HasMany extends Association
 
         if ($ok) {
             $sourceEntity->set($property, $savedEntity->get($property));
-            $sourceEntity->dirty($property, false);
+            $sourceEntity->setDirty($property, false);
         }
 
         return $ok;
@@ -340,6 +374,7 @@ class HasMany extends Association
         $conditions = [
             'OR' => (new Collection($targetEntities))
                 ->map(function ($entity) use ($targetPrimaryKey) {
+                    /** @var \Cake\Datasource\EntityInterface $entity */
                     return $entity->extract($targetPrimaryKey);
                 })
                 ->toList()
@@ -361,7 +396,7 @@ class HasMany extends Association
             );
         }
 
-        $sourceEntity->dirty($property, false);
+        $sourceEntity->setDirty($property, false);
     }
 
     /**
@@ -393,7 +428,7 @@ class HasMany extends Association
      * $author->articles = [$article1, $article2, $article3, $article4];
      * $authors->save($author);
      * $articles = [$article1, $article3];
-     * $authors->association('articles')->replace($author, $articles);
+     * $authors->getAssociation('articles')->replace($author, $articles);
      * ```
      *
      * `$author->get('articles')` will contain only `[$article1, $article3]` at the end
@@ -428,41 +463,43 @@ class HasMany extends Association
      * Deletes/sets null the related objects according to the dependency between source and targets and foreign key nullability
      * Skips deleting records present in $remainingEntities
      *
-     * @param array $properties array of foreignKey properties
+     * @param array $foreignKeyReference The foreign key reference defining the link between the
+     * target entity, and the parent entity.
      * @param \Cake\Datasource\EntityInterface $entity the entity which should have its associated entities unassigned
      * @param \Cake\ORM\Table $target The associated table
      * @param array $remainingEntities Entities that should not be deleted
      * @param array $options list of options accepted by `Table::delete()`
      * @return bool success
      */
-    protected function _unlinkAssociated(array $properties, EntityInterface $entity, Table $target, array $remainingEntities = [], array $options = [])
+    protected function _unlinkAssociated(array $foreignKeyReference, EntityInterface $entity, Table $target, array $remainingEntities = [], array $options = [])
     {
         $primaryKey = (array)$target->getPrimaryKey();
         $exclusions = new Collection($remainingEntities);
         $exclusions = $exclusions->map(
             function ($ent) use ($primaryKey) {
+                /** @var \Cake\Datasource\EntityInterface $ent */
                 return $ent->extract($primaryKey);
             }
         )
         ->filter(
             function ($v) {
-                return !in_array(null, array_values($v), true);
+                return !in_array(null, $v, true);
             }
         )
-        ->toArray();
+        ->toList();
 
-        $conditions = $properties;
+        $conditions = $foreignKeyReference;
 
         if (count($exclusions) > 0) {
             $conditions = [
                 'NOT' => [
                     'OR' => $exclusions
                 ],
-                $properties
+                $foreignKeyReference
             ];
         }
 
-        return $this->_unlink(array_keys($properties), $target, $conditions, $options);
+        return $this->_unlink(array_keys($foreignKeyReference), $target, $conditions, $options);
     }
 
     /**
@@ -598,6 +635,10 @@ class HasMany extends Association
      */
     public function sort($sort = null)
     {
+        deprecationWarning(
+            'HasMany::sort() is deprecated. ' .
+            'Use setSort()/getSort() instead.'
+        );
         if ($sort !== null) {
             $this->setSort($sort);
         }
@@ -637,7 +678,7 @@ class HasMany extends Association
     /**
      * {@inheritDoc}
      *
-     * @return callable
+     * @return \Closure
      */
     public function eagerLoader(array $options)
     {
@@ -655,4 +696,14 @@ class HasMany extends Association
 
         return $loader->buildEagerLoader($options);
     }
+
+    /**
+     * {@inheritDoc}
+     */
+    public function cascadeDelete(EntityInterface $entity, array $options = [])
+    {
+        $helper = new DependentDeleteHelper();
+
+        return $helper->cascadeDelete($this, $entity, $options);
+    }
 }
